(* ::Package:: *)

(* ::Text:: *)
(*Register Functions to Interface with Import[].*)
(*This only stubs the functions with the Mathematica Import/Export framework, so we can use Import to call the converter functions.*)


Begin["ROOTImport`"];


$ROOTImportPath = $InputFileName


If[ StringTake[$Version,1]=!="8",
    ImportExport`RegisterFormat["ROOT", "BinaryFormat" -> True]
]


ImportExport`RegisterImport["ROOT",
{
  (* TTree elements *)
  {"TTreeMetadata", tree_String} :> (ImportTreeMetadata[tree][##]&),
  {"TTreeData", tree_String} :> (ImportTreeData[tree][##]&),
  {"TTreeData", tree_String, leaf_String} :> (ImportBranch[tree, leaf, 0, 0][##]&),

  (* TH1F elements *)
  {"TH1FData", hist_String} :> (ImportTHData[hist, "TH1F"][##]&),
  {"TH1FGraphics", hist_String} :> (ImportTHGraphics[hist, "TH1F"][##]&),

  (* TH2F elements *)
  {"TH2FData", hist_String} :> (ImportTHData[hist, "TH2F"][##]&),
  {"TH2FGraphics", hist_String} :> (ImportTHGraphics[hist, "TH2F"][##]&),

  (* default *)
  ImportKey
},
  "BinaryFormat" -> True,
  "DefaultElement" -> "Keys"
  
  (****************************************************
  Typically we may have a "Sources" option denoting the location where
  the converter functions are defined.  For example,
     "Sources"  -> { "\dir1\file1.m",\dir1\file2.m""}]  }
  When this is not specified, the Import/Export framework automatically
  searches a set of default locations, which include (for the ROOT format)
    { $UserBaseDirectory / SystemFiles / ROOT / Converter.m,
      $UserBaseDirectory / SystemFiles / ROOT / Binaries / $SystemID / ROOT.exe
    }
  *****************************************************)
]


End[];
