#include "TROOT.h"
#include "TCint.h"
#include "TSystem.h"
#include "TInterpreter.h"
#include "TTree.h"
#include "TBranch.h"
#include "TLeaf.h"
#include "TFile.h"
#include "TKey.h"
#include "TH1.h"
#include "TH2F.h"
#include "TAxis.h"
#include <list>
#include <vector>
#include <string>
#include <iterator>

//===================================================================================
//===================================================================================
:Evaluate: Begin["ROOTImport`"];

//===================================================================================
//===================================================================================
:Begin:
:Function: iBuildLibraries
:Pattern: iBuildLibraries[dir_String]
:Arguments: {dir}
:ArgumentTypes: {ByteString}
:ReturnType: Manual
:End:

void iBuildLibraries(kucharp_ct dir, int len1)
{
  gSystem->ChangeDirectory((const char*) dir );
  gSystem->AddDynamicPath( (const char*) dir );
  gInterpreter->GenerateDictionary("vector<vector<int> >",            "vector");
  gInterpreter->GenerateDictionary("vector<vector<unsigned int> >",   "vector");
  gInterpreter->GenerateDictionary("vector<vector<float> >",          "vector");
  gInterpreter->GenerateDictionary("vector<vector<double> >",         "vector");
  gInterpreter->GenerateDictionary("vector<vector<short> >",          "vector");
  gInterpreter->GenerateDictionary("vector<vector<unsigned short> >", "vector");
  gInterpreter->GenerateDictionary("vector<vector<bool> >",           "vector");
  gInterpreter->GenerateDictionary("vector<vector<string> >",         "vector");
  MLPutSymbol(stdlink, "True");
}

:Evaluate: BuildLibraries[] := iBuildLibraries[FileNameJoin[{$UserBaseDirectory,"SystemFiles","Formats","ROOT"}]]

//===================================================================================
//===================================================================================
:Begin:
:Function: iLoadLibraries
:Pattern: iLoadLibraries[dir_String]
:Arguments: {dir}
:ArgumentTypes: {ByteString}
:ReturnType: Manual
:End:

void iLoadLibraries(kucharp_ct dir, int len1)
{
  gSystem->ChangeDirectory((const char*) dir );
  gSystem->AddDynamicPath( (const char*) dir );
  gSystem->Load("AutoDict_vector_vector_int____cxx");
  gSystem->Load("AutoDict_vector_vector_unsigned_int____cxx");
  gSystem->Load("AutoDict_vector_vector_float____cxx");
  gSystem->Load("AutoDict_vector_vector_double____cxx");
  gSystem->Load("AutoDict_vector_vector_short____cxx");
  gSystem->Load("AutoDict_vector_vector_unsigned_short____cxx");
  gSystem->Load("AutoDict_vector_vector_bool____cxx");
  gSystem->Load("AutoDict_vector_vector_string____cxx");
  MLPutSymbol(stdlink, "True");
}

:Evaluate: LoadLibraries[] := iLoadLibraries[FileNameJoin[{$UserBaseDirectory,"SystemFiles","Formats","ROOT"}]]

//===================================================================================
//===================================================================================
:Begin:
:Function: iSetDirectory
:Pattern: iSetDirectory[dir_String]
:Arguments: {dir}
:ArgumentTypes: {ByteString}
:ReturnType: Manual
:End:

void iSetDirectory(kucharp_ct dir, int len1)
{
  gSystem->ChangeDirectory( (const char*) dir );
  gSystem->AddDynamicPath( (const char*) dir );
  MLPutSymbol(stdlink, "True");
}

:Evaluate: SetDir[] := iSetDirectory[FileNameJoin[{$UserBaseDirectory,"SystemFiles","Formats","ROOT"}]]

//===================================================================================
//===================================================================================
:Begin:
:Function: iimportKey
:Pattern: importKey[filename_String]
:Arguments: {filename}
:ArgumentTypes: {ByteString}
:ReturnType: Manual
:End:

void iimportKey(kucharp_ct filename, int len1)
{
  using namespace std;
  
  // adapted from page 163 of the ROOT manual: Chapter 11 
  // since we don't know how many keys there are, we stored them
  // in a vector<string> before transferring to MathLink
  TFile f( (char*) filename);
  TIter next(f.GetListOfKeys());
  TKey *key;

  vector<string> keyList;
  keyList.clear();
  while( (key=(TKey*)next()) )
  {
    keyList.push_back(key->GetName());
    keyList.push_back(key->GetTitle());
    keyList.push_back(key->GetClassName());
  };

  // output in the form of {{Name1, Title1, Class1},{Name2, Title2, Class2}, ...}
  int keyLength = ((int) keyList.size())/3;
  MLPutFunction(stdlink, "List", keyLength );
  for( int i = 0; i!= keyLength; i++ )
  {
    MLPutFunction(stdlink, "List", 3);
    MLPutString(stdlink, (keyList[3*i  ]).c_str() );
    MLPutString(stdlink, (keyList[3*i+1]).c_str() );
    MLPutString(stdlink, (keyList[3*i+2]).c_str() );
  };
  keyList.clear();
  f.Close();
}

//===================================================================================
//===================================================================================
:Begin:
:Function: iimportTreeMetadata
:Pattern: importTreeMetadata[filename_String, tree_String]
:Arguments: {filename, tree}
:ArgumentTypes: {ByteString, ByteString}
:ReturnType: Manual
:End:

void iimportTreeMetadata(kucharp_ct filename, int len1, kucharp_ct tree, int len2)
{
  using namespace std;
  TFile f( (char*) filename);
  TTree *t;
  f.GetObject( (char*) tree, t);
  Long64_t nentries = t->GetEntries();
  TIter next(t->GetListOfLeaves());
  TLeaf *leaf;

  vector<string> out;
  out.clear();
  while( (leaf = (TLeaf*)next()) )
  {
     out.push_back(leaf->GetName());
     out.push_back(leaf->GetTitle());
     out.push_back(leaf->GetTypeName());
  };

  // transfer to mathematica in the form of
  // { {name1, title1, type1, len}, {name2, title2, type2, len}, ... }
  int slen = ((int) out.size())/3;
  MLPutFunction(stdlink, "List", slen);
  for( int i = 0; i!= slen; i++ )
  {
    MLPutFunction(stdlink, "List", 4);
    MLPutString(stdlink, (out[3*i  ]).c_str() );
    MLPutString(stdlink, (out[3*i+1]).c_str() );
    MLPutString(stdlink, (out[3*i+2]).c_str() );
    MLPutInteger(stdlink, (int) nentries );
  };
  out.clear();
  f.Close();
}    

//===================================================================================
//===================================================================================
:Begin:
:Function: iimportTH1F
:Pattern: importTH1F[filename_String, hist_String]
:Arguments: {filename, hist}
:ArgumentTypes: {ByteString, ByteString}
:ReturnType: Manual
:End:

void iimportTH1F(kucharp_ct filename, int len1, kucharp_ct histname, int len2)
{
  using namespace std;
  TFile f( (char*) filename);
  TH1F *hist = new TH1F;
  f.GetObject( (char*) histname, hist);

  TAxis *xaxis = hist->GetXaxis();
  int xbins = xaxis->GetNbins();

  // output in the form of
  // { {bin_lower_edge_1, bin_width_1, bin_content_1, bin_error_1},
  //   {bin_lower_edge_2, bin_width_2, bin_content_2, bin_error_2},
  //   ...}
  MLPutFunction(stdlink, "List", xbins);
  for( int i = 1; i < xbins + 1; ++i )
  {
      MLPutFunction(stdlink, "List", 4);
      MLPutDouble(stdlink, xaxis->GetBinLowEdge(i));
      MLPutDouble(stdlink, xaxis->GetBinWidth(i)  );
      MLPutDouble(stdlink, hist->GetBinContent(i) );
      MLPutDouble(stdlink, hist->GetBinError(i)   );
  };
  delete hist;
  f.Close();
}

//===================================================================================
//===================================================================================
:Begin:
:Function: iimportTH2F
:Pattern: importTH2F[filename_String, hist_String]
:Arguments: {filename, hist}
:ArgumentTypes: {ByteString, ByteString}
:ReturnType: Manual
:End:

void iimportTH2F(kucharp_ct filename, int len1, kucharp_ct histname, int len2)
{
  using namespace std;
  TFile f( (char*) filename);
  TH2F *hist = new TH2F;
  f.GetObject( (char*) histname, hist);

  TAxis *xaxis = hist->GetXaxis();
  TAxis *yaxis = hist->GetYaxis();

  int xbins = xaxis->GetNbins();
  int ybins = yaxis->GetNbins();

  // output in the form of
  // { {x_lower_edge_1, x_width_1, y_lower_edge_1, y_width_1, bin_content_1, bin_error_1},
  //   {x_lower_edge_2, x_width_2, y_lower_edge_2, y_width_2, bin_content_2, bin_error_2},
  //   ...}
  MLPutFunction(stdlink, "List", xbins * ybins);
  for( int i = 1; i < xbins +1; i++){
  for( int j = 1; j < ybins +1; j++)
  {
      MLPutFunction(stdlink, "List", 6);
      MLPutDouble(stdlink, xaxis->GetBinLowEdge(i) );
      MLPutDouble(stdlink, xaxis->GetBinWidth(i)   );
      MLPutDouble(stdlink, yaxis->GetBinLowEdge(j) );
      MLPutDouble(stdlink, yaxis->GetBinWidth(j)   );
      MLPutDouble(stdlink, hist->GetBinContent(i,j));
      MLPutDouble(stdlink, hist->GetBinError(i,j)  );
  }}

  delete hist;
  f.Close();
}

//===================================================================================
//===================================================================================
//*************************************************************************
//********************* MLPutGeneric **************************************
//*************************************************************************

void MLPutGeneric(MLINK &link, const int    &a){ MLPutInteger(link, a);      }
void MLPutGeneric(MLINK &link, const short  &a){ MLPutShortInteger(link, a); }
void MLPutGeneric(MLINK &link, const double &a){ MLPutDouble(link, a);       }
void MLPutGeneric(MLINK &link, const float  &a){ MLPutFloat(link, a);        }
void MLPutGeneric(MLINK &link, char*         a){ MLPutString(link, a );      }
void MLPutGeneric(MLINK &link, const bool   &a){ if(a){ MLPutSymbol(link, "True");} else{ MLPutSymbol(link, "False");} }
void MLPutGeneric(MLINK &link, const std::string &a) { MLPutString(link, a.c_str() );}
void MLPutGeneric(MLINK &link, char &a){ std::string str = &a; std::string str1(str,0,1); MLPutString(link, str1.c_str() ); }
// for unsigned items, upcast to larger holders
void MLPutGeneric(MLINK &link, const unsigned short &a){ MLPutInteger32(link, (int) a);  }
void MLPutGeneric(MLINK &link, const unsigned int &a)
{ 
#if ML64BIT_MATHLINK
   MLPutInteger64(link, (mlint64) a);
#else
   mlint64 t;
   t.hi = 0;
   t.low = a;
   MLPutInteger64(link, t);
#endif
}

// templates have to be defined
template <typename T>
void MLPutGeneric(MLINK &link, std::vector<T> &vec)
{
  int slen = (int) vec.size();
  typename std::vector<T>::iterator it = vec.begin();
  MLPutFunction(link, "List", slen);
  for(it = vec.begin(); it != vec.end(); it++)
  {  
    MLPutGeneric(link, *it);
  }
}

template <typename T>
void MLPutGeneric(MLINK &link, std::list<T> &l)
{
  int slen = (int) l.size();
  typename std::list<T>::iterator it = l.begin();
  MLPutFunction(link, "List", slen);
  for(it = l.begin(); it != l.end(); it++)
  {  
    MLPutGeneric(link, *it);
  }
}

//*********************************************************************************
//************************** BRANCH IMPORTERS *************************************
//*********************************************************************************

// template solution for simple type
template <typename tcpp>
void iGetSimpleBranch(
  char* filename, 
  char* tree,
  char* branch,
  int beg, int end,
  tcpp &vec
)
{
  using namespace std;
  int ibeg(beg), iend(end);

  TFile f(filename);
  TTree *t;
  f.GetObject(tree, t);
  tcpp px;
  TBranch *br = t->GetBranch( branch );
  t->SetBranchAddress(branch, &px, &br);

  Long64_t nentries = t->GetEntries();
  if( iend == 0 ){ iend = (int) nentries; ibeg = 1; }
  MLPutFunction(stdlink, "List", iend-ibeg+1);
  for(int i = (ibeg-1); i<iend; i++)
  {
    long long tentry = t->LoadTree(i); // tentry might be different from i when using a TTree
    br->GetEntry(tentry);
    MLPutGeneric( stdlink, px );
  }
  delete t;
  f.Close();
}


template <>
inline void iGetSimpleBranch<char*>(
  char* filename, 
  char* tree,
  char* branch,
  int beg,
  int end,
  char* &vpchar
)
{
  using namespace std;

  int ibeg(beg), iend(end);
  TFile f(filename);
  TTree *t;
  f.GetObject(tree, t);
  Char_t val[100];
  TBranch *br = t->GetBranch( branch );
  t->SetBranchAddress(branch, val, &br);
  string str;

  Long64_t nentries = t->GetEntries();
  if( iend == 0 ){ iend = (int) nentries; ibeg = 1; }

  MLPutFunction(stdlink, "List", iend-ibeg+1);
  for(int i = (ibeg-1); i<iend; i++)
  {
    long long tentry = t->LoadTree(i); // tentry might be different from i when using a TTree
    br->GetEntry(tentry);
    str = val;
    MLPutString( stdlink, str.c_str() );
  }
  delete t;
  f.Close();
  // val gets destroyed when the function exits; there should be no leak
}


// template solution for complex type: vec<...>
template <typename tcpp>
void iGetComplexBranch(
  char* filename, 
  char* tree,
  char* branch,
  int beg, int end,
  tcpp &vec
)
{
  using namespace std;
  int ibeg(beg), iend(end);

  TFile f(filename);
  TTree *t = 0;
  f.GetObject(tree, t);
  tcpp *px = 0; // this pointer must be initialized!
  TBranch *br = t->GetBranch( branch );
  t->SetBranchAddress(branch, &px, &br);

  Long64_t nentries = t->GetEntries();
  if( iend == 0 ){ iend = (int) nentries; ibeg = 1; }

  MLPutFunction(stdlink, "List", iend-ibeg+1);
  for(int i = (ibeg-1); i<iend; i++)
  {
    long long tentry = t->LoadTree(i); // tentry might be different from i when using a TTree
    br->GetEntry(tentry);
    MLPutGeneric( stdlink, *px );
  }
  delete t;
  delete px;
  f.Close();
}

/******** global buffers *******************************************************/
// The use of global buffers is based on an old design...
// We have to make sure that, when these are used, that they are cleared.

int             sint = 0;
unsigned int    suint = 0;;
float           sfloat = 0.0;
double          sdouble = 0.0;
short           sshort = 0;
unsigned short  sushort = 0;
bool            sbool = true;
std::string     sstring = "string";
char            schar = 'a';
char*           spchar = &schar;

std::vector<int>               vint;
std::vector<unsigned int>      vuint;
std::vector<float>             vfloat;
std::vector<double>            vdouble;
std::vector<short>             vshort;
std::vector<unsigned short>    vushort;
std::vector<bool>              vbool;
std::vector<std::string>       vstring;

std::vector<std::vector<int> >              vvint;
std::vector<std::vector<unsigned int> >     vvuint;
std::vector<std::vector<float> >            vvfloat;
std::vector<std::vector<double> >           vvdouble;
std::vector<std::vector<std::string> >      vvstring;
std::vector<std::vector<short> >            vvshort;
std::vector<std::vector<unsigned short> >   vvushort;
std::vector<std::vector<bool> >             vvbool;

std::list<int>              lint;
std::list<unsigned int>     luint;
std::list<float>            lfloat;
std::list<double>           ldouble;
std::list<short>            lshort;
std::list<unsigned short>   lushort;
std::list<bool>             lbool;
std::list<std::string>      lstring;

//===================================================================================
//===================================================================================
:Begin:
:Function: iImportBranch
:Pattern: importBranch[filename_String, tree_String, br_String, type_Integer, beg_Integer, end_Integer]
:Arguments: {filename, tree, br,  type, beg, end}
:ArgumentTypes: {ByteString, ByteString, ByteString, Integer, Integer, Integer}
:ReturnType: Manual
:End:

void iImportBranch(
  kucharp_ct filename, int len1,
  kucharp_ct tree,     int len2,
  kucharp_ct branch,   int len3,
  int type,
  int beg, int end
)
{
  switch(type)
  {
    /*  simple data types */
    case 1:  iGetSimpleBranch((char*) filename, (char*) tree, (char*) branch, beg, end, sint    ); break;
    case 2:  iGetSimpleBranch((char*) filename, (char*) tree, (char*) branch, beg, end, suint   ); break;
    case 3:  iGetSimpleBranch((char*) filename, (char*) tree, (char*) branch, beg, end, sfloat  ); break;
    case 4:  iGetSimpleBranch((char*) filename, (char*) tree, (char*) branch, beg, end, sdouble ); break;
    case 5:  iGetSimpleBranch((char*) filename, (char*) tree, (char*) branch, beg, end, sshort  ); break;
    case 6:  iGetSimpleBranch((char*) filename, (char*) tree, (char*) branch, beg, end, sushort ); break;
    case 7:  iGetSimpleBranch((char*) filename, (char*) tree, (char*) branch, beg, end, sbool   ); break;
    case 8:  iGetComplexBranch((char*) filename, (char*) tree, (char*) branch, beg, end, sstring); break;
	case 10: iGetSimpleBranch((char*) filename, (char*) tree, (char*) branch, beg, end, spchar  ); break;

    /* vector<*> types */
    case 101: iGetComplexBranch((char*) filename, (char*) tree, (char*) branch, beg, end, vint );    break;
    case 102: iGetComplexBranch((char*) filename, (char*) tree, (char*) branch, beg, end, vuint );   break;
    case 103: iGetComplexBranch((char*) filename, (char*) tree, (char*) branch, beg, end, vfloat );  break;
    case 104: iGetComplexBranch((char*) filename, (char*) tree, (char*) branch, beg, end, vdouble ); break;
    case 105: iGetComplexBranch((char*) filename, (char*) tree, (char*) branch, beg, end, vshort );  break;
    case 106: iGetComplexBranch((char*) filename, (char*) tree, (char*) branch, beg, end, vushort ); break;
    case 107: iGetComplexBranch((char*) filename, (char*) tree, (char*) branch, beg, end, vbool );   break;
    case 108: iGetComplexBranch((char*) filename, (char*) tree, (char*) branch, beg, end, vstring ); break;

    /* vector<vector<*> > types */
    case 201: iGetComplexBranch((char*) filename, (char*) tree, (char*) branch, beg, end, vvint );  break;
    case 202: iGetComplexBranch((char*) filename, (char*) tree, (char*) branch, beg, end, vvuint ); break;
    case 203: iGetComplexBranch((char*) filename, (char*) tree, (char*) branch, beg, end, vvfloat );break;
    case 204: iGetComplexBranch((char*) filename, (char*) tree, (char*) branch, beg, end, vvdouble ); break;
    case 205: iGetComplexBranch((char*) filename, (char*) tree, (char*) branch, beg, end, vvshort ); break;
    case 206: iGetComplexBranch((char*) filename, (char*) tree, (char*) branch, beg, end, vvushort ); break;
    case 207: iGetComplexBranch((char*) filename, (char*) tree, (char*) branch, beg, end, vvbool );  break;
    case 208: iGetComplexBranch((char*) filename, (char*) tree, (char*) branch, beg, end, vvstring ); break;

    /* list<*> types */
    case 301: iGetComplexBranch((char*) filename, (char*) tree, (char*) branch, beg, end, lint );  break;
    case 302: iGetComplexBranch((char*) filename, (char*) tree, (char*) branch, beg, end, luint ); break;
    case 303: iGetComplexBranch((char*) filename, (char*) tree, (char*) branch, beg, end, lfloat ); break;
    case 304: iGetComplexBranch((char*) filename, (char*) tree, (char*) branch, beg, end, ldouble ); break;
    case 305: iGetComplexBranch((char*) filename, (char*) tree, (char*) branch, beg, end, lshort ); break;
    case 306: iGetComplexBranch((char*) filename, (char*) tree, (char*) branch, beg, end, lushort ); break;
    case 307: iGetComplexBranch((char*) filename, (char*) tree, (char*) branch, beg, end, lbool ); break;
    case 308: iGetComplexBranch((char*) filename, (char*) tree, (char*) branch, beg, end, lstring ); break;

    /* should not be here if calls only come from high-level calls */
    default:  MLPutSymbol(stdlink, "$Failed"); break;
  }; 
}

//===================================================================================
//===================================================================================
//__________________________________________________________________________________
//___________________ below is main() -- don't touch! ______________________________
//__________________________________________________________________________________

void InitROOT()
{
  gROOT->ProcessLine("#include <vector>");
  gROOT->ProcessLine("#include <list>");
  gSystem->Load("libRIO");
  gSystem->Load("libTree");
  gSystem->Load("libHist");
  gSystem->Load("libCore");
}

#if MACINTOSH_MATHLINK

int main (int argc, char *argv[])
{
  InitROOT();
  argc = argc; /* suppress warning */
  return MLMain (0, argv);
}

#elif WINDOWS_MATHLINK

#if __BORLANDC__
#pragma argsused
#endif

int PASCAL WinMain (HINSTANCE hInstance,
		HINSTANCE hPrevInstance,
		LPSTR lpszCmdLine,
		int nCmdShow)
{
  InitROOT();  
  char buff[512];
  char FAR *buff_start = buff;
  char FAR *argv[32];
  char FAR *FAR * argv_end = argv + 32;

  MLScanString (argv, &argv_end, &lpszCmdLine, &buff_start);
  return MLMain ((int)(argv_end - argv), argv);
}

#else

int main (int argc, char **argv)
{
  InitROOT();  
  return MLMain (argc, argv);
}

#endif

:Evaluate: End[]
